(ns com.aurellem.run.util
  (:use (com.aurellem.gb util gb-driver vbm characters saves))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(def ↑ [:u])
(def ↓ [:d])
(def ← [:l])
(def → [:r])

(defn first-difference
  [base alt difference-metric [moves root :as script]]
  (loop [branch-point root
         actions moves]
    (let [base-branch (step branch-point base)
          base-val (difference-metric base-branch)
          alt-branch (step branch-point alt)
          alt-val (difference-metric alt-branch)]
      (if (not= base-val alt-val)
        [(conj actions alt) alt-branch]
        (recur base-branch (conj actions base))))))

(defn repeat-until-different
  [buttons metric [moves root :as script]]
  (let [baseline (metric root)]
    (loop [actions (vec moves)
           state root]
      (let [new-state (step state buttons)
            new-actions (conj actions buttons)]
        (if (not= (metric new-state) baseline)
          [new-actions new-state]
          (recur new-actions new-state))))))

(defn delayed-difference
  [base alt delay difference-metric [moves root :as script]]
  (loop [branch-point root
         actions moves]
    (let [base-branch (step branch-point base)
          base-val
          (difference-metric
           (run-moves base-branch
                      (repeat delay base)))
          alt-branch (step branch-point alt)
          alt-val
          (difference-metric
           (run-moves alt-branch
                      (repeat delay base)))]
      (if (not= base-val alt-val)
        [(conj actions alt) alt-branch]
        (recur base-branch (conj actions base))))))


  
;; (defn advance
;;   ([base alt difference-metric [commands state]]
;;      (let [[c s]
;;            (first-difference base alt difference-metric state)]
;;        [(concat commands c) s]))
;;   ([base alt [commands state]]
;;      (advance base alt AF [commands state]))
;;   ([alt [commands state]]
;;      (advance [] alt [commands state])))


(def x-position-address 0xD361)
(def y-position-address 0xD362)

(defn x-position
  ([^SaveState state]
     (aget (memory state) x-position-address))
  ([] (x-position @current-state)))

(defn y-position
  ([^SaveState state]
     (aget (memory state) y-position-address))
  ([] (y-position @current-state)))

(defn move
  [dir script]
  (let [current-position-fn
        (cond (#{← →} dir) x-position
              (#{↑ ↓} dir) y-position)]
    (repeat-until-different dir current-position-fn script)))
  
(defn walk 
  "Move the character along the given directions."
  [directions script]
  (reduce (fn [script dir]
            (move dir script)) script directions))

(defn menu
  [directions script]
  (reduce (fn [script direction]
            (move direction script))
          script directions))


(defn search-string
  [^SaveState state string]
  (let [codes
        (str->character-codes string)
        codes-length (count codes)
        mem (vec (memory state))
        mem-length (count mem)]
    (loop [idx 0]
      (if (< (- mem-length idx) codes-length)
        nil
        (if (= (subvec mem idx (+ idx codes-length))
               codes)
          idx
          (recur (inc idx)))))))

(def text-address 0x9DC1)

(defn displayed-text
  ([^SaveState state]
     (character-codes->str
      (subvec (vec (memory state))
              text-address
              (+ text-address 82))))
  ([] (displayed-text @current-state)))

;; (defn scroll-text
;;   ([script]
;;      (first-difference [:b] [:a :b] AF script))
;;   ([n script]
;;      (reduce (fn [script _]
;;                (scroll-text script))
;;              script
;;              (range n))))

(defn scroll-text
  ([script]
     (delayed-difference
      [:b] [:a :b] 25 displayed-text script))
  ([n script]
     (reduce (fn [script _]
               (scroll-text script))
             script
             (range n))))


(defn end-text [script]
  (->> script
       (scroll-text)
       (play-moves [[] [:a]])))



(common-differences
 (vec (memory (step (talk-to-oak) [:a])))
 (vec (memory (step (talk-to-oak) []))))
 




(defn do-nothing [n script]
  (->> script
       (play-moves
        (repeat n []))))


(defn critical-hit
  "Put the cursor over the desired attack. This program will
   determine the appropriate amount of blank frames to
   insert before pressing [:a] to ensure that the attack is
   a critical hit."
  [script]
  (loop [blanks 6]
    (let [new-script
          (->> script
               (play-moves
                (concat (repeat blanks [])
                        [[:a][]])))]
      (if (let [future-state
                (run-moves (second new-script)
                           (repeat 400 []))

                result (search-string (memory future-state)
                                      "Critical")]
            (if result
              (println "critical hit with" blanks "blank frames"))
            result) 
        new-script
        (recur (inc blanks))))))  

(defn move-thru-grass
  [direction script]
  (loop [blanks 0]
    (let [new-script
          (->> script
               (play-moves (repeat blanks []))
               (move direction))

          future-state
          (run-moves (second new-script)
                     (repeat 600 []))
          
          result (search-string (memory future-state)
                                "Wild")]
      (if (nil? result)
        (do
          (if (< 0 blanks)
            (do
              (println "avoided pokemon with"
                       blanks "blank frames")))
             new-script)
        (recur (inc blanks))))))

(defn walk-thru-grass
  [directions script]
  (reduce (fn [script direction]
            (move-thru-grass direction script))
          script directions))

(defn slowly
  [delay moves script]
  (reduce
   (fn [script move]
     (->> script
          (do-nothing delay)
          (play-moves (vector move))))
   script moves))

(defn multiple-times
  ([n command args script]
     (reduce (fn [script _]
               (apply command (concat args [script])))
             script
             (range n)))
  ([n command script]
     (multiple-times n command [] script)))
